import asyncio
from collections import defaultdict
from collections.abc import Awaitable, Callable, Generator
from typing import Any, Self

from .print import TcrFmt_KwargsDataclass
from .result import Result


def _filter_out_nones_in_place(results: list[Any | None]) -> None:
	results[:] = [r for r in results if r is not None]


def _filter_out_none_results_in_place(results: list[Result[Any | None, Any]]) -> None:
	results[:] = [r for r in results if (r.is_err or r.unwrap() is not None)]


# py313: class BaseEvent[R: Any = None](TcrFmt_KwargsDataclass):
class BaseEvent[R: Any](TcrFmt_KwargsDataclass):
	"""Represents a base event, kind of like BaseException represents base exception. Generic on the return value."""

	_subscribers = defaultdict(set)

	@classmethod
	def subscribe(cls, func: Callable[[Self], Awaitable[R]]):
		"""Subscribe to an event by passing a function shaped as: `async (Event) -> Event<R>::R`."""
		if not asyncio.iscoroutinefunction(func):
			raise TypeError("Subscriber must be an async function")
		cls._subscribers[cls].add(func)
		return func

	def _iter_subscriber_calls(self) -> Generator[Callable[[Self], Awaitable[R]]]:
		for cls in type(self).mro():
			if not issubclass(cls, BaseEvent):
				continue

			yield from self._subscribers.get(cls, set())

	async def emit_results(self, *, filter_out_nones: bool = False) -> list[Result[R, Exception]]:
		"""Emit the event instance, calling & awaiting all relevant subscribers. Returns a tuple of all subscriber return values (if subscriber raises, appropriate Result value is used in the given position)."""
		results = []
		for func in self._iter_subscriber_calls():
			try:
				current_returned = await func(self)
			except Exception as e:
				results.append(Result.new_err(e))
			else:
				results.append(Result.new_ok(current_returned))

		if filter_out_nones:
			_filter_out_none_results_in_place(results)

		return results

	async def emit_excgroup(self, *, filter_out_nones: bool = False) -> list[R]:
		"""Emit the event instance, calling & awaiting all relevant subscribers. Returns a tuple of all subscriber return values (if subscriber raises, an exception group is raised, emitting continues, and at the end the accumulated exceptions in the exception group are raised)."""
		results = []
		exceptions = []

		for func in self._iter_subscriber_calls():
			try:
				results.append(await func(self))
			except Exception as e:
				exceptions.append(e)

		if exceptions:
			raise ExceptionGroup("During event emission at least 1 subscriber raised an exception", exceptions)

		if filter_out_nones:
			_filter_out_nones_in_place(results)

		return results

	async def emit_failfast(self, *, filter_out_nones: bool = False) -> list[R]:
		"""Emit the event instance, calling & awaiting all relevant subscribers. Returns a tuple of all subscriber return values (if subscriber raises, dont catch the error, the emission ends immediately)."""
		results = []

		for func in self._iter_subscriber_calls():
			results.append(await func(self))

		if filter_out_nones:
			_filter_out_nones_in_place(results)

		return results


if __name__ == "__main__":

	async def __main():  # Basic inline testing
		import pydantic as pd

		from .console import c

		class BM(pd.BaseModel):
			model_config = pd.ConfigDict(
				arbitrary_types_allowed=True,
				validate_assignment=True,
			)

		class UserEvent(BaseEvent[str]):
			def __init__(self, username: str):
				self.username = username

		class UserLoginEvent(UserEvent):
			def __init__(self, username: str):
				super().__init__(username)

		class PydanticEvent(BM, BaseEvent[str]):
			userid: int

		####################

		@UserLoginEvent.subscribe
		async def on_user_login(event: UserLoginEvent):
			print(f"User {event.username} logged in.")
			return "Login Handled"

		@UserEvent.subscribe
		async def on_any_user_event(event: UserEvent):
			print(f"User event detected: {event!r}")
			return "User Event Processed"

		######

		class BadEvent(BaseEvent): ...

		@BadEvent.subscribe
		async def bad_callback(event: BadEvent):
			raise ValueError("whyvent")

		@BadEvent.subscribe
		async def ok_callback(event: BadEvent):
			return "ok"

		for _ in range(10):

			@BadEvent.subscribe
			async def bad_callback2(event: BadEvent):
				raise RuntimeError("evąt")

		#######

		@PydanticEvent.subscribe
		async def on_pydantic_event(event: PydanticEvent):
			print(f"Pydantic event detected: {event.userid!r}")
			return "Pydantic Event Processed"

		if True:  # happy path
			event = UserLoginEvent("Alice")
			results = await event.emit_excgroup()
			c("Subscribers returned:", results)

			c.hr()

			event = UserLoginEvent("Alice")
			results = await event.emit_results()
			c("Subscribers returned:", results)

			c.hr()

			event = UserLoginEvent("Alice")
			results = await event.emit_failfast()
			c("Subscribers returned:", results)

			c.hr()

		if True:  # Pydantic happy path
			event = PydanticEvent(userid=42)
			results = await event.emit_excgroup()
			c("Subscribers returned:", results)

		if True:  # Printing
			c(UserLoginEvent("alice"))
			c(PydanticEvent(userid=42))
			c(UserLoginEvent)
			c(PydanticEvent)

		if False:  # bad path
			event = BadEvent()
			results = await event.emit_results()
			c("Bad Subscribers returned:", results)

			c.hr()

			event = BadEvent()
			results = await event.emit_excgroup()
			c("Bad Subscribers returned:", results)

			c.hr()

	asyncio.run(__main())



from datetime import date, datetime, time
from datetime import timedelta as Δ
from datetime import tzinfo as _TzInfo
from functools import partial
from types import EllipsisType

import pytest
from tcrutils.console import c
from tcrutils.decorator import copy_kwargs
from tcrutils.timestr2 import timestr as t

tz = datetime.now().astimezone().tzinfo


if True:  # Fixtures

	@pytest.fixture()
	def true_now():
		return now_tz_μ0()


if True:  # Helpers

	def now_tz_μ0():
		return datetime.now(tz=tz).replace(microsecond=0)

	def eq_datetime(
		a: datetime,
		b: datetime,
		*,
		tolerance: Δ = Δ(seconds=0),
		print_=True,
	):
		if print_:
			print()
			c("a=", a)
			c("b=", b)

		diff = abs(a - b)

		if print_:
			c("diff=", diff)

		result = diff <= tolerance

		if not result:
			c.error("not result!")

		return result

	def datetime_dict(dt: datetime):
		return {
			"year": dt.year,
			"month": dt.month,
			"day": dt.day,
			"hour": dt.hour,
			"minute": dt.minute,
			"second": dt.second,
			"microsecond": dt.microsecond,
			"tz": dt.tzinfo,
		}

	@copy_kwargs
	def mdtn_factory(
		*,
		year: int | EllipsisType = ...,
		month: int | EllipsisType = ...,
		day: int | EllipsisType = ...,
		hour: int | EllipsisType = ...,
		minute: int | EllipsisType = ...,
		second: int | EllipsisType = ...,
		microsecond: int | EllipsisType = ...,
		tz: _TzInfo | None = tz,
		offset: Δ | None = Δ(0),
		__kwargs: dict[str, int] = ...,
	):
		if "tz" in __kwargs:
			del __kwargs["tz"]

		if "offset" in __kwargs:
			del __kwargs["offset"]

		if offset is None:
			offset = Δ(0)

		def mock_datetime_now(tz: _TzInfo = tz, __offset=offset):
			return datetime.now(tz=tz).replace(**__kwargs) + __offset

		return mock_datetime_now, mock_datetime_now()


@pytest.mark.parametrize(
	("m",),
	(
		(1,),
		(2,),
		(3,),
		(7,),
		(10,),
	),
)
def test_robostr_all_units(true_now: datetime, *, m: int):
	mdtn, mock_now = mdtn_factory(**datetime_dict(true_now))

	parse = partial(t.parse, tz=tz, _datetime_now=mdtn)

	if now_1s := mock_now + Δ(seconds=1) * m:
		assert eq_datetime(parse(f"{m}s"), now_1s)
		assert eq_datetime(parse(f"{m}sec"), now_1s)
		assert eq_datetime(parse(f"{m}secs"), now_1s)
		assert eq_datetime(parse(f"{m}sex"), now_1s)
		assert eq_datetime(parse(f"{m}second"), now_1s)
		assert eq_datetime(parse(f"{m}seconds"), now_1s)

	if now_1m := mock_now + Δ(minutes=1) * m:
		assert eq_datetime(parse(f"{m}m"), now_1m)
		assert eq_datetime(parse(f"{m}min"), now_1m)
		assert eq_datetime(parse(f"{m}mins"), now_1m)
		assert eq_datetime(parse(f"{m}minute"), now_1m)
		assert eq_datetime(parse(f"{m}minutes"), now_1m)

	if now_1h := mock_now + Δ(hours=1) * m:
		assert eq_datetime(parse(f"{m}h"), now_1h)
		assert eq_datetime(parse(f"{m}hr"), now_1h)
		assert eq_datetime(parse(f"{m}hrs"), now_1h)
		assert eq_datetime(parse(f"{m}hour"), now_1h)
		assert eq_datetime(parse(f"{m}hours"), now_1h)

	if now_1d := mock_now + Δ(days=1) * m:
		assert eq_datetime(parse(f"{m}d"), now_1d)
		assert eq_datetime(parse(f"{m}day"), now_1d)
		assert eq_datetime(parse(f"{m}days"), now_1d)

	if now_1w := mock_now + Δ(days=7) * m:
		assert eq_datetime(parse(f"{m}w"), now_1w)
		assert eq_datetime(parse(f"{m}week"), now_1w)
		assert eq_datetime(parse(f"{m}weeks"), now_1w)

	if now_1y := mock_now + Δ(days=365) * m:
		assert eq_datetime(parse(f"{m}y"), now_1y)
		assert eq_datetime(parse(f"{m}year"), now_1y)
		assert eq_datetime(parse(f"{m}years"), now_1y)

	if now_1decade := mock_now + Δ(days=365 * 10) * m:
		assert eq_datetime(parse(f"{m}decade"), now_1decade)
		assert eq_datetime(parse(f"{m}decades"), now_1decade)

	if now_1rescue := mock_now + Δ(hours=6) * m:
		assert eq_datetime(parse(f"{m}res"), now_1rescue)
		assert eq_datetime(parse(f"{m}reses"), now_1rescue)
		assert eq_datetime(parse(f"{m}resees"), now_1rescue)
		assert eq_datetime(parse(f"{m}rescue"), now_1rescue)
		assert eq_datetime(parse(f"{m}rescues"), now_1rescue)

	if now_1pull := mock_now + Δ(hours=11, minutes=30) * m:
		assert eq_datetime(parse(f"{m}pul"), now_1pull)
		assert eq_datetime(parse(f"{m}pull"), now_1pull)
		assert eq_datetime(parse(f"{m}puls"), now_1pull)
		assert eq_datetime(parse(f"{m}pulls"), now_1pull)
		assert eq_datetime(parse(f"{m}card"), now_1pull)
		assert eq_datetime(parse(f"{m}cards"), now_1pull)


def test_robostr_multi_subsegment(true_now: datetime):
	mdtn, mock_now = mdtn_factory(**datetime_dict(true_now))

	for s, δ in (
		("1h15m", Δ(hours=1, minutes=15)),
		("2h-4m", Δ(hours=2, minutes=-4)),
		("-.5h", Δ(hours=-0.5)),
	):
		assert eq_datetime(
			t.parse(s, tz=tz, _datetime_now=mdtn),
			mock_now + δ,
		)


@pytest.mark.parametrize(
	("weekday_str", "weekday_num"),
	(
		("mon", 0),
		("tue", 1),
		("wed", 2),
		("thu", 3),
		("fri", 4),
		("sat", 5),
		("sun", 6),
	),
)
def test_weekday_with_mock(true_now: datetime, *, weekday_str: str, weekday_num: int):
	mdtn, mock_now = mdtn_factory(
		hour=true_now.hour,
		minute=0,
		second=1,
		microsecond=0,
		tz=tz,
	)

	calculated = mock_now + Δ(days=1)

	while calculated.weekday() != weekday_num:
		calculated = calculated + Δ(days=1)

	assert eq_datetime(t.parse(weekday_str, tz=tz, _datetime_now=mdtn), calculated)


_test_time_with_mock_parametrize_body = [
	( None, None, None, Δ(0)       ),
	( None, 0,    1,    Δ(0)       ),
	( None, 0,    0,    Δ(0)       ),
	( 13,   None, None, Δ(0)       ),
	( 13,   0,    1,    Δ(0)       ),
	( 13,   0,    0,    Δ(0)       ),
	( 21,   None, None, Δ(0)       ),
	( 21,   0,    1,    Δ(0)       ),
	( 21,   0,    0,    Δ(0)       ),
	( None, None, None, Δ(days=1)  ),
	( None, 0,    1,    Δ(days=1)  ),
	( None, None, None, Δ(days=-1) ),
	( None, 0,    1,    Δ(days=-1) ),
]  # fmt: skip


@pytest.mark.parametrize(("h", "m", "s", "offset"), _test_time_with_mock_parametrize_body)
def test_time_with_mock(true_now: datetime, *, h: int | None, m: int | None, s: int | None, offset: Δ):
	mdtn, mock_now = mdtn_factory(
		hour=h if h is not None else true_now.hour,
		minute=m if m is not None else true_now.minute,
		second=s if s is not None else true_now.second,
		microsecond=0,
		offset=offset,
		tz=tz,
	)

	for st, hh, mm, ss in [
		(f"{mock_now.hour}:", mock_now.hour, 0, 0),
		(f"{mock_now.hour}::", mock_now.hour, 0, 0),
		(f"{mock_now.hour}:6:9", mock_now.hour, 6, 9),
	]:
		assert eq_datetime(
			t.parse(
				st,
				_datetime_now=mdtn,
			),
			mock_now.replace(minute=mm, second=ss) + Δ(days=int(mock_now.timetz() >= time(hour=hh, minute=mm, second=ss, tzinfo=tz))),
		)



from collections.abc import Callable

from .extract_error import extract_error, extract_traceback
from .iterable import cut_at

BACKTICK = "`"
BACKTICKS = 3 * BACKTICK
NEWLINE = "\n"


def codeblock(
	text: str,
	*,
	langcode: str = "",
	max_length: int = 1984,
	cut_at_func: Callable[[str, int], str] = cut_at,
	smart_empty: bool = True,
) -> str:
	"""Pack text into a Discord codeblock.

	Note: May return str with len > max_length if max_length is ridicously low such that not even `'''\\nlangcode'''` fits

	Args:
	  text: The text to put in the codeblock body.
	  langcode: The language code to use in discord codeblock, for example "py".
	  max_length: The maximum length of the output str.
	  cut_at_func: A Callable[[str, int], str] which's returned string is of or less than the passed in int in length.
	  smart_empty: If true, if `text` is '' (an empty string) it will be changed to ' ' (a space). This prevents the codeblock to think langcode is the body of the codeblock if no text is supplied.
	"""
	if smart_empty and text == "":
		text = " "

	maxlen = (
		max_length - (len(BACKTICKS) + len(langcode) + len(NEWLINE) + len(BACKTICKS)) if max_length != -1 else 9999999999999999  # Eh...
	)
	return BACKTICKS + langcode + NEWLINE + cut_at_func(text, maxlen) + BACKTICKS


def uncodeblock(text: str) -> str:
	"""*Try to* remove any discord codeblocks along with their langcodes if any are found."""
	if text[-3:] == BACKTICKS and text[:3] == BACKTICKS:
		code_start = 3
		code_end = -3
		if NEWLINE in text[3:]:  # Check if there is a language code specified
			code_start = text.index(NEWLINE) + 1
		return text[code_start:code_end].strip()
	return text


def codeblocks(
	text1: str,
	*texts: str,
	langcodes: tuple[str] | None = None,
	max_length: int = 1984,
	cut_at_func: Callable[[str, int], str] = cut_at,
	smart_empty: bool = True,
):
	"""Mutliple-codeblock(). May suffer simillar issues as `codeblock`, like the total_max_length not actually being max because codeblock() may return something slightly longer.

	For more info read codeblock() docstring
	"""
	texts = text1, *texts

	if langcodes is None:
		langcodes = ("",) * len(texts)

	out = ""

	for text, langcode in zip(texts, langcodes, strict=True):
		out += codeblock(text, langcode=langcode, max_length=max_length - len(out), cut_at_func=cut_at_func, smart_empty=smart_empty)

	return out


def discord_exception(e: BaseException, *, max_length: int = 1800) -> str:
	return codeblocks(
		extract_error(e),
		extract_traceback(e),
		langcodes=(
			"txt",
			"py",
		),
		max_length=max_length,
		smart_empty=True,
	)



from collections.abc import Callable, Coroutine
from dataclasses import dataclass
from enum import Enum, auto
from functools import wraps
from typing import Any, Generic, Literal, Self, TypeVar, overload

from .print import FMT_BRACKETS


class EmptyType(Enum):
	EMPTY = auto()


Default = TypeVar("Default")


class ResultUnwrappedOnErrorError(Exception):
	"""An unwrap() was called on a Result containing an error."""


class ResultUnwrappedErrOnValueError(Exception):
	"""An unwrap_err() was called on a Result contaning a value."""


class _ResultBase[Ok, Err: BaseException]:
	_value: Ok | Err
	_is_err: bool

	@overload
	def __init__(self, *, is_error: Literal[False], value: Ok): ...

	@overload
	def __init__(self, *, is_error: Literal[True], value: Err): ...

	def __init__(
		self,
		*,
		is_error: bool = False,
		value: Ok | Err,
	):
		self._is_err = bool(is_error)

		if is_error and not isinstance(value, BaseException):
			raise ValueError(f"If is_error=True, value must be an {'**INSTANCE**' if isinstance(value, type) else 'instance'} of a subclass of BaseException")

		self._value = value

	@classmethod
	def new_ok(cls, value: Ok):
		return cls(is_error=False, value=value)

	@classmethod
	def new_err(cls, error: Err):
		return cls(is_error=True, value=error)

	@property
	def is_ok(self) -> bool:
		return not self._is_err

	@property
	def is_err(self) -> bool:
		return self._is_err

	def __eq__(self, other):
		if not isinstance(other, self.__class__):
			return NotImplemented

		return self._is_err == other._is_err and self._value == other._value

	def unwrap_err(self) -> Err:
		"""If the result contains a value, raise `ResultUnwrappedErrOnValueError()`, else return the error."""
		if not self._is_err:
			raise ResultUnwrappedErrOnValueError() from self._value

		return self._value

	def unwrap_or(self, default: Default) -> Ok | Default:
		"""If the result contains an error, return default, else return the value."""
		if self.is_err:
			return default

		return self._value

	def unwrap_err_or(self, default: Default) -> Err | Default:
		"""If the result contains a value, return default, else return the error."""
		if not self.is_err:
			return default

		return self._value

	def unwrap_or_else(self, default_factory: Callable[[Err], Default]) -> Ok | Default:
		"""If the result contains an error, return `default_factory(that_error)`, else return the value."""
		if self.is_err:
			return default_factory(self._value)

		return self._value

	def unwrap_err_or_else(self, default_factory: Callable[[Ok], Default]) -> Err | Default:
		"""If the result contains a value, return `default_factory(that_value)`, else return the error."""
		if not self.is_err:
			return default_factory(self._value)

		return self._value

	def raise_if_possible(self) -> None:
		"""If the result contains a value, do nothing, otherwise raise the error."""

		if self.is_err:
			raise self._value

	def map_err(self, f: Callable[[Err], Err]) -> Self:
		"""If the result contains an error, set it to `f(current_error)`, else do nothing."""
		if self.is_err:
			self._value = f(self._value)

		return self

	def map_ok(self, f: Callable[[Ok], Ok]) -> Self:
		"""If the result contains a value (Ok), set it to `f(current_value)`, else do nothing."""
		if self.is_ok:
			self._value = f(self._value)

		return self

	def __tcr_fmt__(self=None, *, fmt_iterable, syntax_highlighting, **kwargs):
		if self is None:
			raise NotImplementedError

		return (
			fmt_iterable(self.__class__)
			+ fmt_iterable(696969.696969).replace("696969", "")
			+ fmt_iterable(bool(self.is_ok)).replace("True", "Ok").replace("False", "Err")
			+ (FMT_BRACKETS[tuple][syntax_highlighting] % fmt_iterable(self._value))
		)

	def unwrap_direct(self) -> Ok:
		"""If the result contains an error, raise it, else return the value."""
		if self._is_err:
			raise self._value

		return self._value

	def unwrap_indirect(self) -> Ok:
		"""If the result contains an error, raise `ResultUnwrappedOnErrorError()`, else return the value."""
		if self._is_err:
			raise ResultUnwrappedOnErrorError() from self._value

		return self._value


class Result[Ok, Err: BaseException](_ResultBase[Ok, Err]):
	"""May contain a value or an error, but not both. Check with `.is_ok` or `.unwrap()`, etc."""

	def unwrap(self) -> Ok:
		"""If the result contains an error, raise `ResultUnwrappedOnErrorError()`, else return the value (On `Result` alias to `.unwrap_indirect()`)."""
		return self.unwrap_indirect()


class Result2[Ok, Err: BaseException](Result[Ok, Err]):
	"""Like Result, but .unwrap() in the sad case raises the underling error, not `ResultUnwrappedOnErrorError()` (On `Result2` alias to `.unwrap_direct()`). Normal `Result`'s `.unwrap()` can also be accessed via `.unwrap_indirect()`."""

	def unwrap(self) -> Ok:
		"""If the result contains an error, raise it, else return the value."""
		return self.unwrap_direct()


def aresultify[**P, R](f: Callable[P, Coroutine[Any, Any, R]]) -> Callable[P, Coroutine[Any, Any, Result[R, Exception]]]:
	@wraps(f)
	async def wrapper(*args, **kwargs):
		try:
			return Result.new_ok(await f(*args, **kwargs))
		except Exception as e:
			return Result.new_err(e)

	return wrapper


def resultify[**P, R](f: Callable[P, R]) -> Callable[P, Result[R, Exception]]:
	@wraps(f)
	def wrapper(*args, **kwargs):
		try:
			return Result.new_ok(f(*args, **kwargs))
		except Exception as e:
			return Result.new_err(e)

	return wrapper


def aresultify2[**P, R](f: Callable[P, Coroutine[Any, Any, R]]) -> Callable[P, Coroutine[Any, Any, Result2[R, Exception]]]:
	@wraps(f)
	async def wrapper(*args, **kwargs):
		try:
			return Result2.new_ok(await f(*args, **kwargs))
		except Exception as e:
			return Result2.new_err(e)

	return wrapper


def resultify2[**P, R](f: Callable[P, R]) -> Callable[P, Result2[R, Exception]]:
	@wraps(f)
	def wrapper(*args, **kwargs):
		try:
			return Result2.new_ok(f(*args, **kwargs))
		except Exception as e:
			return Result2.new_err(e)

	return wrapper



"""Not to be confused with `tcrutils.language`. This module is for language translation helper(s)."""

from collections.abc import Callable, Hashable
from typing import overload

from .print import tcrfmt_dataclass_by_name_args_kwargs as _tcrfmt_dataclass_by_name_args_kwargs


# py313: class Lang[LocaleK: Hashable = str, TranslationK: Hashable = str, V = str]:
class Lang[LocaleK: Hashable, TranslationK: Hashable, V]:
	_d: dict[LocaleK | None, dict[TranslationK, V]]
	_merge_func: Callable[[V, V], V]

	def __init__(self, merge_func: Callable[[V, V], V] = lambda v1, v2: v2):  # converter: Callable[[LocaleK, TranslationK, V], V] = lambda v: v
		self._d = {}
		self._merge_func = merge_func

	def __repr__(self):
		return f"{self.__class__.__name__}({self._d!r})"

	def __tcr_fmt__(self=None, *, fmt_iterable, syntax_highlighting, **kwargs):
		if self is None:
			raise NotImplementedError

		return _tcrfmt_dataclass_by_name_args_kwargs(
			fmt_iterable=fmt_iterable,
			name=self.__class__.__name__,
			syntax_highlighting=syntax_highlighting,
			obj_kwargs=self._d,
		)

	def _merge_dicts(self, old: dict[TranslationK, V], new: dict[TranslationK, V]) -> dict[TranslationK, V]:
		merged = {}

		all_keys = set(old) | set(new)

		for k in all_keys:
			if k in old and k in new:
				merged[k] = self._merge_func(old[k], new[k])
			elif k in old:
				merged[k] = old[k]
			else:
				merged[k] = new[k]

		return merged

	def merge_localepack(self, locale: LocaleK | None, pack: dict[TranslationK, V]) -> None:
		if locale is not None:
			self._d[None] = pack | self._d.get(None, {})

		self._d[locale] = self._merge_dicts(self._d.get(locale, {}), pack)

	def getitem_localepack(self, locale: LocaleK | None) -> dict[TranslationK, V]:
		return self._d[locale]

	def get_localepack(self, locale: LocaleK | None, default: dict[TranslationK, V] | None = None) -> dict[TranslationK, V] | None:
		return self._d.get(locale, default)

	@overload
	def __getitem__(self, key: LocaleK | None) -> dict[TranslationK, V]: ...

	@overload
	def __getitem__(self, key: tuple[LocaleK | None, TranslationK]) -> V: ...

	@overload
	def __getitem__(self, key: tuple[slice, TranslationK]) -> dict[LocaleK, V]: ...

	def __getitem__(self, key: LocaleK | None) -> dict[TranslationK, V]:
		if not isinstance(key, tuple):
			return self._d[key]

		if not len(key) == 2:
			raise TypeError("if tuple, key must be of length 2")

		k1, k2 = key

		if isinstance(k1, slice):
			if (k1.start, k1.stop, k1.step) != (None, None, None):
				raise ValueError('No compound slices supported, only ":"')

			return {locale: self._d[locale][k2] for locale in self._d if locale is not None if k2 in self._d[locale]}

		return self._d[k1][k2]

	def get_hikari(self, key: TranslationK, *, name: str | None = None) -> dict[str, V | dict[LocaleK, V]]:
		"""Handle hikari (arc) localizations for the given key, return splattable (`**result`) dict with keys: `{name}` and `{name}_localizations` with the appropriate values.

		Args:
			name: If not specified (None), the str(key) will be split by "." and the last part will be used. Warning: this may lead to unexpected results, if the key is not a str itself or does not have a well-made __str__
		"""
		if name is None:
			name = str(key).split(".")[-1]

		localized = self[:, key]

		unlocalized = self[None, key]

		return {
			f"{name}": unlocalized,
			f"{name}_localizations": localized,
		}

	def get_arc_command(self, key: TranslationK) -> dict[str, V | dict[LocaleK, V]]:
		"""Handle hikari-arc's `@slash_command` localizations, by providing a dict with appripriate keys to splat (**result).

		## `!!!` This will only work if the TranslationK generic is str, since it then fetches a value with concatenated str as a key
		"""
		return self.get_hikari(f"{key!s}.name") | self.get_hikari(f"{key!s}.description")


if __name__ == "__main__":

	def __main():
		from hikari import Locale as L

		from .console import c

		LOCALE_INDEP = {
			"hi.name": "hi default name",
			"hi.desc": "hi default desc",
			"bye": "bye default",
		}

		PL = {
			"hi.name": "Witaj",
			"hi.desc": "hej słodziaku~",
			"bye": "Do widzenia",
		}

		EN_GB = {
			"hi.name": "Greetings",
			"hi.desc": "Hi cutie~",
			"hi.extra": "UwU",
			"bye": "Farewell",
		}

		lang = Lang[L]()

		lang.merge_localepack(None, LOCALE_INDEP)

		lang.merge_localepack(L.PL, PL)
		lang.merge_localepack(L.EN_GB, EN_GB)
		lang.merge_localepack(L.EN_GB, {"nya": "nyaa"})

		c(lang[:, "hi.extra"])
		c(lang)
		c("lang[:, 'hi.name']=", eval=True)
		c("lang[:, 'hi.desc']=", eval=True)
		c("lang[L.PL, 'hi.desc']=", eval=True)
		c('lang.get_hikari("hi.name", name="name")=', eval=True)
		c('lang.get_hikari("hi.name")=', eval=True)

		c.hr()

		LOCALE_INDEP2 = {
			"hello": ("Hi", "hai~", "Hello"),
			"bye": ("Bye", "Goodbye"),
		}

		PL2 = {
			"hello": ("Witaj", "hej"),
			"bye": ("Do widzenia", "Do widzenya~"),
		}

		EN_GB2 = {
			"hello": ("Hi", "Hello"),
			"bye": ("Bye", "Goodbye"),
		}

		pools = Lang[L, str, tuple[str]](merge_func=lambda v1, v2: (*v1, *v2))

		pools.merge_localepack(None, LOCALE_INDEP2)

		pools.merge_localepack(L.PL, PL2)
		pools.merge_localepack(L.EN_GB, EN_GB2)
		pools.merge_localepack(L.EN_GB, {"hello": ("hewwo",)})

		c(pools)
		c("pools[:, 'hello']=", eval=True)
		c("pools[:, 'bye']=", eval=True)
		c("pools[L.PL, 'bye']=", eval=True)
		c("pools.get_hikari('hello', name='hello')=", eval=True)

	__main()



def test_batched():
	from tcrutils.iterable import batched

	assert batched("1234567890", n=3) == ["123", "456", "789", "0"]
	assert batched("1234567890", n=3, back_to_front=True) == ["1", "234", "567", "890"]
	assert batched("", n=3) == []


def test_cut_at():
	from tcrutils.iterable import cut_at

	assert cut_at("uwuwuwuwuwu", n=10) == "uwuwuwu..."
	assert cut_at("uwuwuwuwuw", n=10) == "uwuwuwuwuw"
	assert cut_at("https://google.com/path", n=100, shrink_links_visually_if_fits=True) == "[https:/google.com/](<https://google.com/path>)"
	assert cut_at("https://google.com/path", n=20, shrink_links_visually_if_fits=True) == "https://google.co..."



import inspect
import os
import pathlib as p
import re
from collections.abc import Callable


def convert_to_relative_path(
	absolute_path: str | p.Path,
	relative_to: str | p.Path | None = None,
) -> p.Path:
	"""Convert absolute path into a relative path, if not possible, return the absolute path."""
	relative_to = relative_to if relative_to is not None else p.Path.cwd()

	absolute_path = p.Path(absolute_path)

	try:
		relative_path = absolute_path.relative_to(relative_to)
	except ValueError:
		return str(absolute_path)

	return p.Path() / relative_path


def convert_to_relative_pathstr(
	absolute_path: str | p.Path,
	relative_to: str | p.Path | None = None,
	*,
	consistent_slashes: bool = True,
) -> str:
	"""Convert absolute path into a relative path string, if not possible, return the absolute path string."""
	converted = convert_to_relative_path(absolute_path=absolute_path, relative_to=relative_to)
	converted = str(converted)

	if consistent_slashes and os.name == "nt":
		converted = converted.replace("\\", "/")

	if any(converted.startswith(x) for x in ("\\", "/")):
		return converted

	if os.name == "nt" and ":" in converted:
		return converted

	return f".{'\\' if (os.name == 'nt' and not consistent_slashes) else '/'}{converted}"


def nextname(name="New folder") -> str:
	"""### If passed in name does not end with "`({int})" where `{int}` is any integer, return it with "` (1)`" appended else return it with the previously mentioned `{int}` incremented and leave it otherwise unchanged."""
	match = re.match(r"^(.*) \((\d+)\)$", name)

	if match:
		base_name = match.group(1)
		current_index = int(match.group(2))
		new_name = f"{base_name} ({current_index + 1})"
	else:
		new_name = f"{name} (1)"

	return new_name


def nextname_file_ext_fix(name="New file.txt") -> str:
	match = re.match(r"^(.*?)(?:\s*\((\d+)\))?(?:\.(\w+))?$", name)
	if not match:
		raise ValueError("Invalid filename format")

	base_name, number, extension = match.groups()

	new_number = int(number) + 1 if number else 1

	return f"{base_name} ({new_number}).{extension}" if extension else f"{base_name} ({new_number})"


def get_unused_dirname(
	name: str | None = None,
	path: p.Path | str | None = None,
	nextname: Callable[[str | None], str] = nextname,
) -> str:
	"""Return unused directory or file with the name `name` in directory `path` or the current directory if not passed in."""
	if path is None:
		path = os.getcwd()
	path = str(path)
	listdir = os.listdir(path)
	while name in listdir:
		name = nextname(name)
	return name


def pyrootfile() -> p.Path:
	"""Return the file from which the python code execution originated, furthest back in the callstack as pathlib.Path."""
	return p.Path(inspect.stack()[-1].filename)


def pyrootdir() -> p.Path:
	"""Return the directory of the furthest back file in the callstack as pathlib.Path."""
	return pyrootfile().parent


if True:  # isreserved
	nt_reserved_chars = frozenset({chr(i) for i in range(32)} | {'"', "*", ":", "<", ">", "?", "|", "/", "\\"})

	nt_reserved_names = frozenset({"CON", "PRN", "AUX", "NUL", "CONIN$", "CONOUT$"} | {f"COM{c}" for c in "123456789\xb9\xb2\xb3"} | {f"LPT{c}" for c in "123456789\xb9\xb2\xb3"})

	def isreserved(path: str | p.Path, *, strict: bool = True) -> bool:
		"""Return true if the pathname is reserved by the system.

		Part backport of Py3.13's os.path.isreserved() and also works on Linux (probably).

		Args:
			strict: If true, mark some more paths as reserved on Windows, on Linux no effect.
		"""

		path = str(path)

		if os.name != "nt":
			return "\x00" in path

		if strict:
			if path.startswith(" "):
				return True

		from os.path import altsep, sep, splitroot

		# Refer to "Naming Files, Paths, and Namespaces":
		# https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
		path = os.fsdecode(splitroot(path)[2]).replace(altsep, sep)
		return any(_isreservedname(name) for name in reversed(path.split(sep)))

	def _isreservedname(name):
		"""Return true if the filename is reserved by the system."""
		# Trailing dots and spaces are reserved.
		if name[-1:] in (".", " "):
			return name not in (".", "..")
		# Wildcards, separators, colon, and pipe (*?"<>/\:|) are reserved.
		# ASCII control characters (0-31) are reserved.
		# Colon is reserved for file streams (e.g. "name:stream[:type]").
		if nt_reserved_chars.intersection(name):
			return True
		# DOS device names are reserved (e.g. "nul" or "nul .txt"). The rules
		# are complex and vary across Windows versions. On the side of
		# caution, return True for names that may not be reserved.
		return name.partition(".")[0].rstrip(" ").upper() in nt_reserved_names



import fnmatch

from ..console import console as c
from . import nodes as m_nodes


def get_nodes_by_name_from_root_nodes(
	root_nodes: tuple["m_nodes.Node"],
	name: str,
	*names: str,
) -> tuple["m_nodes.Node"]:
	maching_nodes = []

	for node in root_nodes:
		if isinstance(node, str):
			# c.warn("Ignoring isinstance(node, str)==True in get_nodes_by_name_from_root_nodes")
			continue

		if not fnmatch.fnmatch(node.name, name):
			continue

		maching_nodes.append(node)

	if not names:
		return tuple(maching_nodes)

	return get_nodes_by_name_from_root_nodes([child for node in maching_nodes for child in node.children], *names)


def parse_and_submit_nodes(
	input_str: str,
	nodes: tuple["m_nodes.Node | str", ...],
	_root_nodes: tuple["m_nodes.Node", ...] = None,
) -> tuple["m_nodes.Node"]:
	if _root_nodes is None:
		_root_nodes = nodes  # No need to copy i tihnk?

	for node in nodes:
		groups = node.match(input_str)

		if groups is None:
			continue

		if len(groups) == 2:
			groups = (*groups, False)

		node_text, rest, incomplete = groups

		node = node.copy()

		node.submit(node_text)

		node_children = [
			z
			for y in (
				(
					# get_nodes_by_name_from_root_nodes(
					# 	_root_nodes,
					# 	*x.split("/"),
					# )
					node.resolve_path(x) if isinstance(x, str) else (x,)
				)
				for x in node.children
			)
			for z in y
		]

		if rest:
			rest = parse_and_submit_nodes(rest, node_children, _root_nodes)
		elif any(node_children_irrefutable := [x for x in node_children if isinstance(x, m_nodes.IrrefutableNode)]):
			rest = parse_and_submit_nodes(rest, node_children_irrefutable, ())
		else:
			rest = ()

		if incomplete:
			incomplete = (m_nodes.IncompleteNode(unknown_text="?"),)
		else:
			incomplete = ()

		return (node, *incomplete, *rest)

	return (m_nodes.UnknownNode(unknown_text=input_str),)
